/*
 * $Id: timer.c,v 1.6 1998/07/17 18:40:47 masaki Exp $
 */

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <sys/time.h>
#include <stdarg.h>
#include <defs.h>
#include <mrt.h>
#include <timer.h>
#include <version.h>


/* #define DEBUG_TIMER	1 */
 
Timer_Master *TIMER_MASTER;


/* timer_jitter
 * return some random number based on timer jitter value and srand
 */
int static timer_jitter (mtimer_t *timer) {
  int val;
  
  if (timer->jitter <= 0) {val =  0;}
  else {
    val = (rand() % (timer->jitter * 2 + 1)) - timer->jitter;
  }
  
  return (val);
}


/* left
 * return the number of seconds left before timer will fire
 */
long time_left (mtimer_t *timer) {
  if (timer == NULL) {return (-1);}

  return (timer->time_next_fire - time (NULL));
}
  

/* Timer_Compare
 * Compare two timers for use in sorting. A timer that is 
 * ON always comes before OFF timers
 */
int Timer_Compare (mtimer_t *t1, mtimer_t *t2) {

#ifdef DEBUG_TIMER
   printf("SORTING: Comparing %s:%d/%d and %s:%d/%d\n",
	  t1->name, t1->time_next_fire, t1->on,
	  t2->name, t2->time_next_fire, t2->on);
#endif
   
   if ((t1->on == 0) && (t2->on != 0))
      return (1);
   if ((t2->on == 0) && (t1->on != 0))
      return (-1);

   return (t1->time_next_fire - t2->time_next_fire);
}

 
sigset_t
block_signal (int sig)
{
   sigset_t new, old;

   sigemptyset (&new);
   sigaddset (&new, sig);
#ifdef HAVE_LIBPTHREAD
   pthread_sigmask (SIG_BLOCK, &new, &old);
#else
   sigprocmask (SIG_BLOCK, &new, &old);
#endif /* HAVE_LIBPTHREAD */
   return (old);
}


void
recover_signal (sigset_t old)
{
#ifdef HAVE_LIBPTHREAD
   pthread_sigmask (SIG_SETMASK, &old, NULL);
#else
   sigprocmask (SIG_SETMASK, &old, NULL);
#endif /* HAVE_LIBPTHREAD */
}


/* Timer_Master_Fire
 * Multiplex alarm signal to list of timers. Fire timers
 * if next fire time before current time
 */
void Timer_Master_Fire () {
   mtimer_t *p_timer;
   time_t now;
   time_t time_next_fire;
   int changed = 0;
   LINKED_LIST *ll_tmp;

#ifdef DEBUG_TIMER
   printf("Master TIMER FIRE!\n");
#endif

   pthread_mutex_lock (&TIMER_MASTER->mutex_lock);

   assert (LL_GetCount (TIMER_MASTER->ll_timers));

   ll_tmp = LL_Create (0);

   time (&now);
   LL_Iterate (TIMER_MASTER->ll_timers, p_timer) {

      if (p_timer->on == 0 ||
	  p_timer->time_next_fire > now)
	 continue;

      LL_Add (ll_tmp, p_timer);
      changed++;

      if (p_timer->flags & TIMER_ONE_SHOT) {
          p_timer->time_next_fire = 0;
   	  p_timer->on = 0;
      }
      else {
	 p_timer->time_interval = 
	    p_timer->time_interval_base + timer_jitter (p_timer);
	 assert (p_timer->time_interval > 0);
	 p_timer->time_next_fire += p_timer->time_interval;
      }
   }

   if (changed)
      LL_Sort (TIMER_MASTER->ll_timers);

   if ((p_timer = LL_GetHead (TIMER_MASTER->ll_timers)) && p_timer->on) {
#ifdef DEBUG_TIMER
   printf("Head of list = %s:%d\n", p_timer->name, p_timer->time_next_fire);
#endif
      TIMER_MASTER->time_interval = p_timer->time_interval;
      TIMER_MASTER->time_next_fire = p_timer->time_next_fire;
   }
   else {
#ifdef DEBUG_TIMER
      if (p_timer)
         printf("Head of list = %s:%d but OFF\n", p_timer->name, 
		p_timer->time_next_fire);
#endif
      TIMER_MASTER->time_interval = 10;        /* always fire just in case */
      TIMER_MASTER->time_next_fire = now + 10; /* always fire just in case */
   }

   /* save it to be changed by others */
   time_next_fire = TIMER_MASTER->time_next_fire;

   pthread_mutex_unlock (&TIMER_MASTER->mutex_lock);

   /* okay, finally schedule timer calls */
   LL_Iterate (ll_tmp, p_timer) {
     
     trace (TR_TIMER | TR_TRACE, TIMER_MASTER->trace, "Timer_Master_Fire: %s\n", 
        p_timer->name);
     p_timer->call_fn (p_timer, p_timer->arg);
   }
		 
   LL_Destroy (ll_tmp);

   /* next alarm set here because we don't want to be called before returning */
   if (time_next_fire > 0) {
      time_next_fire -= time (NULL);
      if (time_next_fire <= 0)
	 time_next_fire = 1;
      alarm (time_next_fire);
   }

#ifdef DEBUG_TIMER
   printf("Master TIMER set to %d\n", time_next_fire);
   if (time_next_fire <= 1) {
     while (1); /* loop so gdb can break here */
   }
#endif
}


static int alarm_pending = 0;

/* check to see if alarm is pending */
void
mrt_alarm (void)
{
#ifdef HAVE_LIBPTHREAD
    int sig;
    sigset_t set;

    sigemptyset (&set);
    sigaddset (&set, SIGALRM);

    /* wait on alarm */
    if (sigwait (&set, &sig) < 0) {
       trace (TR_TIMER | TR_TRACE, TIMER_MASTER->trace, 
	  "sigwait: %s\n", strerror (errno));
    }
    else {
	if (sig == SIGALRM)
	    Timer_Master_Fire ();
    }
#else
    if (alarm_pending) {
	alarm_pending = 0;
	Timer_Master_Fire ();
    }
#endif /* HAVE_LIBPTHREAD */
}


/* This is the only function that's invoked by asynchronous signal.
   most functions are not asynchronous signal safe, so don't call
   them during interrupt. Timers MRT uses are second order so that
   they don't require rapid real-time processing */
static void 
alarm_interrupt (void)
{
    alarm_pending++;
}


/* init_timer
 * Allocate memory and initialize global TIMER_MASTER variable
 */
void init_timer (trace_t *tr) {

  sigset_t set;

  assert (TIMER_MASTER == NULL);

  TIMER_MASTER = New (Timer_Master);
  
  TIMER_MASTER->ll_timers = LL_Create (LL_CompareFunction, Timer_Compare,
			       LL_AutoSort, True, 0);
  TIMER_MASTER->time_interval = 0;
  TIMER_MASTER->time_next_fire = 0;
  TIMER_MASTER->trace = tr;

#ifdef HAVE_SIGACTION /* POSIX actually */
  {
       struct sigaction act;

       memset (&act, 0, sizeof (act));
       act.sa_handler = alarm_interrupt;
#ifdef SA_RESTART
       act.sa_flags = SA_RESTART; 
#endif
       sigaction (SIGALRM, &act, NULL);
  }
#else
#ifdef HAVE_SIGSET
   sigset (SIGALRM, alarm_interrupt);
#else
   signal (SIGALRM, alarm_interrupt);
#endif /* HAVE_SIGSET */
#endif /* POSIX signals */

   pthread_mutex_init (&TIMER_MASTER->mutex_lock, NULL);

   /* This init routine must be called from the main thread
      that handles alarm interrupts */
    sigemptyset (&set);
    sigaddset (&set, SIGALRM);
#ifdef HAVE_LIBPTHREAD
    /* sigwait() will be used so that ALARM must be blocked */
    pthread_sigmask (SIG_BLOCK, &set, NULL);
#else
    /* enables alarm interrupts */
    pthread_sigmask (SIG_UNBLOCK, &set, NULL);
#endif /* HAVE_LIBPTHREAD */
}


/* Destroy_Timer_Master
 */
void Destroy_Timer_Master (Timer_Master *timer) {
   LL_Destroy (timer->ll_timers);
   Destroy (timer);
}


/* Timer_Turn_OFF
 * Turn a timer off and update the TIMER_MASTER info
 */
void Timer_Turn_OFF (mtimer_t *timer) {
   if (timer == NULL) return;
   
   timer->time_next_fire = 0;
   timer->on = 0;
   
   Timer_Master_Update (timer);
   trace (TR_TIMER | TR_TRACE, TIMER_MASTER->trace, 
	  "Timer_Turn_OFF: %s\n", timer->name);
}


/* Timer_Turn_ON
 * Turn a timer ON and update TIMER_MASTER info
 */
void Timer_Turn_ON (mtimer_t *timer) {
   timer->time_interval = timer->time_interval_base + timer_jitter (timer);
   assert (timer->time_interval > 0);
   timer->time_next_fire = time(NULL) + timer->time_interval;
   timer->on = 1;

   Timer_Master_Update (timer);
   trace (TR_TIMER | TR_TRACE, TIMER_MASTER->trace, "Timer_Turn_ON: %s with %d\n", 
      timer->name, timer->time_interval);
}


/* Timer_Reset_Time
 * Reset timer to fire in timer->time_interval and 
 * notify TIMER_MASTER of change
 */

void Timer_Reset_Time (mtimer_t *timer) {
   timer->time_interval = timer->time_interval_base + timer_jitter (timer);
   assert (timer->time_interval > 0);
   timer->time_next_fire = time(NULL) + timer->time_interval;
   timer->on = 1;

   if (timer->time_interval <= 0) 
      timer->on = 0;

   Timer_Master_Update (timer);
   trace (TR_TIMER | TR_TRACE, TIMER_MASTER->trace, "Timer_Reset_Time: %s with %d \n", 
      timer->name, timer->time_interval);
}


/* Timer_Set_Time
 * Change intreval of timer and notify 
 * TIMER_MASTER of new info.
 */
void Timer_Set_Time (mtimer_t *timer, int interval) {
   timer->time_interval_base = interval;
   timer->time_interval = timer->time_interval_base + timer_jitter (timer);
   assert (timer->time_interval > 0);
   timer->time_next_fire = time(NULL) + timer->time_interval;
   timer->on = 1;

   if (timer->time_interval <= 0) 
      timer->on = 0;

   Timer_Master_Update (timer);
   trace (TR_TRACE | TR_TIMER, TIMER_MASTER->trace, "Timer_Set_Time: %s with %d \n", 
      timer->name, timer->time_interval);
} 

void timer_set_flags (mtimer_t *timer, int flags) {

   timer->flags = flags;
} 


int timer_set_jitter (mtimer_t *timer, int jitter) {
  timer->jitter = jitter;
  assert (jitter >= 0);
  return 0;
}


/* Timer_Master_Update
 * Update global TIMER_MASTER structure with new, 
 * or changed timer. Resort timer list and
 * change next_fire if timer fire before, and
 * set alarm.
 */
void Timer_Master_Update (mtimer_t *p_timer) {

   pthread_mutex_lock (&TIMER_MASTER->mutex_lock);

   LL_ReSort (TIMER_MASTER->ll_timers, p_timer);

   if ((p_timer = LL_GetHead (TIMER_MASTER->ll_timers)) == NULL) {
      pthread_mutex_unlock (&TIMER_MASTER->mutex_lock);
      return;
   }
   
   if (p_timer->on == 0) {
      if (TIMER_MASTER->time_next_fire) {
         TIMER_MASTER->time_interval = 0;
         TIMER_MASTER->time_next_fire = 0;
         alarm (0);
#ifdef DEBUG_TIME
         printf("Master mtimer_t has been updated OFF\n");
#endif
      }
      pthread_mutex_unlock (&TIMER_MASTER->mutex_lock);
      return;
   }

   if ((TIMER_MASTER->time_next_fire == 0) ||
       (TIMER_MASTER->time_next_fire >  p_timer->time_next_fire)) {
      time_t t;
      
      TIMER_MASTER->time_next_fire = p_timer->time_next_fire;
      TIMER_MASTER->time_interval = p_timer->time_interval;

      t = TIMER_MASTER->time_next_fire - time (NULL);
      if (t <= 0) t = 1;
      alarm (t);
#ifdef DEBUG_TIMER
      printf("Master mtimer_t has been updated to %d\n", 
	     TIMER_MASTER->time_interval);
#endif
   }

   pthread_mutex_unlock (&TIMER_MASTER->mutex_lock);
}



/* New_Timer
 * Allocate memory and initialize a mtimer_t structure. 
 * Also add new mtimer_t info to global TIMER_MASTER
 * NOTE: timers are created OFF!
 */
mtimer_t *New_Timer (void (*call_fn)(), int interval, char *name, void *arg) {
   mtimer_t *tmp;

   tmp = New (mtimer_t);
   
   tmp->call_fn = call_fn;
   tmp->arg = arg;
   tmp->time_interval_base = interval;
   tmp->time_interval = 0;
   tmp->time_next_fire = 0; /* later calculate */
   tmp->jitter = 0;
   tmp->on = 0;
   if (name != NULL)
      tmp->name = (char *) strdup (name);
   else
      tmp->name = strdup ("unknown");

   pthread_mutex_lock (&TIMER_MASTER->mutex_lock);

   LL_Add (TIMER_MASTER->ll_timers, tmp);

   pthread_mutex_unlock (&TIMER_MASTER->mutex_lock);

   Timer_Master_Update (tmp);


   return (tmp);
}


void
Destroy_Timer (mtimer_t *timer)
{
    Timer_Turn_OFF (timer);
    Timer_Master_Update (timer);
    pthread_mutex_lock (&TIMER_MASTER->mutex_lock);
    LL_Remove (TIMER_MASTER->ll_timers, timer);
    pthread_mutex_unlock (&TIMER_MASTER->mutex_lock);

    Delete (timer->name);
    Delete (timer);
}


